from enum import IntFlag

import gbd_mapping
import numpy as np
import pandas as pd
import pytest
from gbd_mapping import ModelableEntity, causes, covariates, risk_factors

from tests.extract.check import RUNNING_ON_CI
from vivarium_inputs import core, utility_data
from vivarium_inputs.mapping_extension import healthcare_entities

pytestmark = pytest.mark.skipif(
    RUNNING_ON_CI, reason="Don't run these tests on the CI server"
)


def success_expected(entity_name, measure_name, location):
    df = core.get_data(entity_name, measure_name, location)
    return df


def fail_expected(entity_name, measure_name, location):
    with pytest.raises(Exception):
        df = core.get_data(entity_name, measure_name, location)


class MCFlag(IntFlag):
    """
    Use the idea of a bit field with support from python's enum type IntFlag
    See here for general information on bit fields:
    https://en.wikipedia.org/wiki/Bit_field

    And here for python's enum type:
    https://docs.python.org/3.6/library/enum.html
    """

    INCIDENCE_RATE = 1
    RAW_INCIDENCE_RATE = 2
    PREVALENCE = 4
    BIRTH_PREVALENCE = 8
    DISABILITY_WEIGHT = 16
    REMISSION_RATE = 32
    CAUSE_SPECIFIC_MORTALITY_RATE = 64
    EXCESS_MORTALITY_RATE = 128
    DEATHS = 512


entity = [
    (
        causes.measles,
        MCFlag.INCIDENCE_RATE
        | MCFlag.RAW_INCIDENCE_RATE
        | MCFlag.PREVALENCE
        | MCFlag.DISABILITY_WEIGHT
        | MCFlag.CAUSE_SPECIFIC_MORTALITY_RATE
        | MCFlag.EXCESS_MORTALITY_RATE
        | MCFlag.DEATHS,
    ),
    (
        causes.diarrheal_diseases,
        MCFlag.INCIDENCE_RATE
        | MCFlag.RAW_INCIDENCE_RATE
        | MCFlag.PREVALENCE
        | MCFlag.DISABILITY_WEIGHT
        | MCFlag.REMISSION_RATE
        | MCFlag.CAUSE_SPECIFIC_MORTALITY_RATE
        | MCFlag.EXCESS_MORTALITY_RATE
        | MCFlag.DEATHS,
    ),
    (
        causes.diabetes_mellitus_type_2,
        MCFlag.INCIDENCE_RATE
        | MCFlag.RAW_INCIDENCE_RATE
        | MCFlag.PREVALENCE
        | MCFlag.DISABILITY_WEIGHT
        | MCFlag.CAUSE_SPECIFIC_MORTALITY_RATE
        | MCFlag.EXCESS_MORTALITY_RATE
        | MCFlag.DEATHS,
    ),
]
measures = [
    ("incidence_rate", MCFlag.INCIDENCE_RATE),
    ("raw_incidence_rate", MCFlag.RAW_INCIDENCE_RATE),
    ("prevalence", MCFlag.PREVALENCE),
    ("birth_prevalence", MCFlag.BIRTH_PREVALENCE),
    ("disability_weight", MCFlag.DISABILITY_WEIGHT),
    ("remission_rate", MCFlag.REMISSION_RATE),
    ("cause_specific_mortality_rate", MCFlag.CAUSE_SPECIFIC_MORTALITY_RATE),
    ("excess_mortality_rate", MCFlag.EXCESS_MORTALITY_RATE),
    ("deaths", MCFlag.DEATHS),
]
locations = ["India"]


@pytest.mark.parametrize("entity", entity, ids=lambda x: x[0].name)
@pytest.mark.parametrize("measure", measures, ids=lambda x: x[0])
@pytest.mark.parametrize("location", locations)
def test_core_causelike(entity, measure, location):
    entity_name, entity_expected_measure_ids = entity
    measure_name, measure_id = measure
    tester = success_expected if (entity_expected_measure_ids & measure_id) else fail_expected
    df = tester(entity_name, measure_name, utility_data.get_location_id(location))


class MRFlag(IntFlag):
    """
    Use the idea of a bit field with support from python's enum type IntFlag
    See here for general information on bit fields:
    https://en.wikipedia.org/wiki/Bit_field

    And here for python's enum type:
    https://docs.python.org/3.6/library/enum.html
    """

    EXPOSURE = 1
    EXPOSURE_SD = 2
    EXPOSURE_DIST_WEIGHTS = 4
    RELATIVE_RISK = 8
    PAF = 16
    # not implemented
    # MEDIATION_FACTORS = 32
    ALL = EXPOSURE | EXPOSURE_SD | EXPOSURE_DIST_WEIGHTS | RELATIVE_RISK | PAF


entity_r = [
    (
        risk_factors.high_systolic_blood_pressure,
        MRFlag.EXPOSURE
        | MRFlag.EXPOSURE_SD
        | MRFlag.EXPOSURE_DIST_WEIGHTS
        | MRFlag.RELATIVE_RISK
        | MRFlag.PAF,
    ),
    (
        risk_factors.low_birth_weight_and_short_gestation,
        MRFlag.EXPOSURE | MRFlag.RELATIVE_RISK | MRFlag.PAF,
    ),
]
measures_r = [
    ("exposure", MRFlag.EXPOSURE),
    ("exposure_standard_deviation", MRFlag.EXPOSURE_SD),
    ("exposure_distribution_weights", MRFlag.EXPOSURE_DIST_WEIGHTS),
    ("relative_risk", MRFlag.RELATIVE_RISK),
    ("population_attributable_fraction", MRFlag.PAF),
]
locations_r = ["India"]


@pytest.mark.parametrize("entity", entity_r, ids=lambda x: x[0].name)
@pytest.mark.parametrize("measure", measures_r, ids=lambda x: x[0])
@pytest.mark.parametrize("location", locations_r)
def test_core_risklike(entity, measure, location):
    entity_name, entity_expected_measure_ids = entity
    measure_name, measure_id = measure
    tester = success_expected if (entity_expected_measure_ids & measure_id) else fail_expected
    df = tester(entity_name, measure_name, utility_data.get_location_id(location))


entity_cov = [
    covariates.systolic_blood_pressure_mmhg,
]
measures_cov = ["estimate"]
locations_cov = ["India"]


@pytest.mark.parametrize("entity", entity_cov, ids=lambda x: x.name)
@pytest.mark.parametrize("measure", measures_cov, ids=lambda x: x)
@pytest.mark.parametrize("location", locations_cov)
def test_core_covariatelike(entity, measure, location):
    df = core.get_data(entity, measure, utility_data.get_location_id(location))


@pytest.mark.parametrize(
    "measures",
    [
        "structure",
        "age_bins",
        "demographic_dimensions",
        "theoretical_minimum_risk_life_expectancy",
    ],
)
def test_core_population(measures):
    pop = ModelableEntity("ignored", "population", None)
    df = core.get_data(pop, measures, utility_data.get_location_id("India"))


# TODO - Underlying problem with gbd access. Remove when corrected.
entity_health_system = [
    healthcare_entities.outpatient_visits,
]
measures_health_system = ["utilization_rate"]
locations_health_system = ["India"]


@pytest.mark.skip(reason="Underlying problem with gbd access. Remove when corrected.")
@pytest.mark.parametrize("entity", entity_health_system, ids=lambda x: x.name)
@pytest.mark.parametrize("measure", measures_health_system, ids=lambda x: x)
@pytest.mark.parametrize("location", locations_health_system)
def test_core_healthsystem(entity, measure, location):
    df = core.get_data(entity, measure, utility_data.get_location_id(location))



from typing import List

import pandas as pd
from gbd_mapping import RiskFactor

from vivarium_inputs.globals import NON_MAX_TMREL, SEXES, gbd


def get_estimation_years(*_, **__) -> pd.Series:
    data = gbd.get_estimation_years()
    return data


def get_year_block(*_, **__) -> pd.DataFrame:
    estimation_years = get_estimation_years()
    year_block = pd.DataFrame(
        {"year_start": range(min(estimation_years), max(estimation_years) + 1)}
    )
    year_block["year_end"] = year_block["year_start"] + 1
    return year_block


def get_age_group_ids(*_, **__) -> List[int]:
    data = gbd.get_age_group_id()
    return data


def get_age_bins(*_, **__) -> pd.DataFrame:
    age_bins = gbd.get_age_bins()[
        ["age_group_id", "age_group_name", "age_group_years_start", "age_group_years_end"]
    ].rename(columns={"age_group_years_start": "age_start", "age_group_years_end": "age_end"})
    return age_bins


def get_location_id(location_name):
    return {r.location_name: r.location_id for _, r in gbd.get_location_ids().iterrows()}[
        location_name
    ]


def get_location_id_parents(location_id: int) -> List[int]:
    location_metadata = gbd.get_location_path_to_global().set_index("location_id")
    parent_ids = [
        int(loc) for loc in location_metadata.at[location_id, "path_to_top_parent"].split(",")
    ]
    return parent_ids


def get_demographic_dimensions(
    location_id: int, draws: bool = False, value: float = None
) -> pd.DataFrame:
    ages = get_age_group_ids()
    estimation_years = get_estimation_years()
    years = range(min(estimation_years), max(estimation_years) + 1)
    sexes = [SEXES["Male"], SEXES["Female"]]
    location = [location_id]
    values = [location, sexes, ages, years]
    names = ["location_id", "sex_id", "age_group_id", "year_id"]

    data = pd.MultiIndex.from_product(values, names=names).to_frame(index=False)
    if draws:
        for i in range(1000):
            data[f"draw_{i}"] = value
    return data


def get_tmrel_category(entity: RiskFactor) -> str:
    if entity.name in NON_MAX_TMREL:
        tmrel_cat = NON_MAX_TMREL[entity.name]
    else:
        tmrel_cat = sorted(list(entity.categories.to_dict()), key=lambda x: int(x[3:]))[-1]
    return tmrel_cat



import logging

import pytest
from _pytest.logging import caplog as _caplog
from gbd_mapping import causes, etiologies, risk_factors
from loguru import logger


def pytest_addoption(parser):
    parser.addoption("--runslow", action="store_true", default=False, help="run slow tests")


def pytest_configure(config):
    config.addinivalue_line("markers", "slow: mark test as slow to run")


def pytest_collection_modifyitems(config, items):
    if config.getoption("--runslow"):
        # --runslow given in cli: do not skip slow tests
        return
    skip_slow = pytest.mark.skip(reason="need --runslow option to run")
    for item in items:
        if "slow" in item.keywords:
            item.add_marker(skip_slow)


@pytest.fixture
def caplog(_caplog):
    class PropogateHandler(logging.Handler):
        def emit(self, record):
            logging.getLogger(record.name).handle(record)

    handler_id = logger.add(PropogateHandler(), format="{message}")
    yield _caplog
    logger.remove(handler_id)


@pytest.fixture
def cause_list():
    return [
        causes.diarrheal_diseases,
        causes.ischemic_heart_disease,
        causes.ischemic_stroke,
        causes.hemorrhagic_stroke,
        causes.tetanus,
        causes.diabetes_mellitus,
        causes.all_causes,
    ]


@pytest.fixture
def etiology_list():
    return [etiologies.cholera, etiologies.amoebiasis]


@pytest.fixture
def sequela_list():
    return list(
        causes.diarrheal_diseases.sequelae
        + causes.ischemic_heart_disease.sequelae
        + causes.ischemic_stroke.sequelae
        + causes.hemorrhagic_stroke.sequelae
        + causes.hemorrhagic_stroke.sequelae
        + causes.tetanus.sequelae
        + causes.diabetes_mellitus.sequelae
    )


@pytest.fixture
def etiology_list():
    return list(
        causes.diarrheal_diseases.etiologies + causes.lower_respiratory_infections.etiologies
    )


@pytest.fixture
def risk_list():
    return [r for r in risk_factors]


@pytest.fixture
def locations():
    return ["Bangladesh", "Ethiopia", "Kenya", "China", "North Korea", "Nigeria"]



from typing import Union

from gbd_mapping.base_template import GbdRecord, ModelableEntity
from gbd_mapping.id import me_id


class HealthcareEntity(ModelableEntity):
    """Container for healthcare system GBD ids and data."""

    __slots__ = ("name", "kind", "gbd_id", "utilization")

    def __init__(
        self,
        name: str,
        kind: str,
        gbd_id: Union[me_id, None],
        utilization: me_id = None,
    ):
        super().__init__(name=name, kind=kind, gbd_id=gbd_id)
        self.utilization = utilization


class HealthcareEntities(GbdRecord):
    """Holder of healthcare modelable entities"""

    __slots__ = ("outpatient_visits", "inpatient_visits")

    def __init__(
        self, outpatient_visits: HealthcareEntity, inpatient_visits: HealthcareEntity
    ):
        super().__init__()
        self.outpatient_visits = outpatient_visits
        self.inpatient_visits = inpatient_visits



from typing import Union

import pandas as pd
from gbd_mapping import Cause, Covariate, Etiology, RiskFactor, Sequela

import vivarium_inputs.validation.raw as validation
from vivarium_inputs.globals import (
    MEASURES,
    METRICS,
    OTHER_MEID,
    DataAbnormalError,
    DataDoesNotExistError,
    EmptyDataFrameException,
    InputsException,
    NoBestVersionError,
    Population,
    gbd,
)
from vivarium_inputs.mapping_extension import AlternativeRiskFactor, HealthcareEntity
from vivarium_inputs.utilities import filter_to_most_detailed_causes


def extract_data(
    entity, measure: str, location_id: int, validate: bool = True, **get_draws_kwargs
) -> Union[pd.Series, pd.DataFrame]:
    """Check metadata for the requested entity-measure pair. Pull raw data from
    GBD. The only filtering that occurs is by applicable measure id, metric id,
    or to most detailed causes where relevant. If validate is turned on, will
    also pull any additional data needed for raw validation and call raw
    validation on the extracted data.

    Parameters
    ----------
    entity
        Entity for which to extract data.
    measure
        Measure for which to extract data.
    location_id
        Location for which to extract data.
    validate
        Flag indicating whether additional data needed for raw validation
        should be extracted and whether raw validation should be performed.
        Should only be set to False if data is being extracted for
        investigation. Never extract data for a simulation without validation.

    Returns
    -------
    Data for the entity-measure pair and specific location requested.

    Raises
    ------
    DataDoesNotExistError
        If data for the entity-measure-location set requested does not exist.

    """
    extractors = {
        # Cause-like measures
        "incidence_rate": (extract_incidence_rate, {}),
        "prevalence": (extract_prevalence, {}),
        "birth_prevalence": (extract_birth_prevalence, {}),
        "disability_weight": (extract_disability_weight, {}),
        "remission_rate": (extract_remission_rate, {}),
        "deaths": (extract_deaths, {"population": extract_structure}),
        # Risk-like measures
        "exposure": (extract_exposure, {}),
        "exposure_standard_deviation": (
            extract_exposure_standard_deviation,
            {"exposure": extract_exposure},
        ),
        "exposure_distribution_weights": (extract_exposure_distribution_weights, {}),
        "relative_risk": (extract_relative_risk, {"exposure": extract_exposure}),
        "population_attributable_fraction": (
            extract_population_attributable_fraction,
            {"exposure": extract_exposure, "relative_risk": extract_relative_risk},
        ),
        "etiology_population_attributable_fraction": (
            extract_population_attributable_fraction,
            {},
        ),
        "mediation_factors": (extract_mediation_factors, {}),
        # Covariate measures
        "estimate": (extract_estimate, {}),
        # Health system measures
        "utilization_rate": (extract_utilization_rate, {}),
        # Population measures
        "structure": (extract_structure, {}),
        "theoretical_minimum_risk_life_expectancy": (
            extract_theoretical_minimum_risk_life_expectancy,
            {},
        ),
    }

    validation.check_metadata(entity, measure)

    try:
        main_extractor, additional_extractors = extractors[measure]
        data = main_extractor(entity, location_id, **get_draws_kwargs)
    except (
        ValueError,
        AssertionError,
        EmptyDataFrameException,
        NoBestVersionError,
        InputsException,
    ) as e:
        if isinstance(
            e, ValueError
        ) and f"Metadata associated with rei_id = {entity.gbd_id}" not in str(e):
            raise e
        elif (
            isinstance(e, AssertionError)
            and f"Invalid covariate_id {entity.gbd_id}" not in str(e)
            and "No best model found" not in str(e)
        ):
            raise e
        elif isinstance(e, InputsException) and measure != "birth_prevalence":
            raise e
        else:
            raise DataDoesNotExistError(
                f"{measure.capitalize()} data for {entity.name} does not exist."
            )

    if validate:
        additional_data = {
            name: extractor(entity, location_id)
            for name, extractor in additional_extractors.items()
        }
        validation.validate_raw_data(data, entity, measure, location_id, **additional_data)

    return data


def extract_prevalence(entity: Union[Cause, Sequela], location_id: int) -> pd.DataFrame:
    data = gbd.get_incidence_prevalence(
        entity_id=entity.gbd_id, location_id=location_id, entity_type=entity.kind
    )
    data = data[data.measure_id == MEASURES["Prevalence"]]
    return data


def extract_incidence_rate(entity: Union[Cause, Sequela], location_id: int) -> pd.DataFrame:
    data = gbd.get_incidence_prevalence(
        entity_id=entity.gbd_id, location_id=location_id, entity_type=entity.kind
    )
    data = data[data.measure_id == MEASURES["Incidence rate"]]
    return data


def extract_birth_prevalence(entity: Union[Cause, Sequela], location_id: int) -> pd.DataFrame:
    data = gbd.get_birth_prevalence(
        entity_id=entity.gbd_id, location_id=location_id, entity_type=entity.kind
    )
    data = data[data.measure_id == MEASURES["Incidence rate"]]
    return data


def extract_remission_rate(entity: Cause, location_id: int) -> pd.DataFrame:
    data = gbd.get_modelable_entity_draws(entity.me_id, location_id)
    data = data[data.measure_id == MEASURES["Remission rate"]]
    return data


def extract_disability_weight(entity: Sequela, location_id: int) -> pd.DataFrame:
    disability_weights = gbd.get_auxiliary_data(
        "disability_weight", entity.kind, "all", location_id
    )
    data = disability_weights.loc[
        disability_weights.healthstate_id == entity.healthstate.gbd_id, :
    ]
    return data


def extract_deaths(entity: Cause, location_id: int) -> pd.DataFrame:
    data = gbd.get_codcorrect_draws(entity.gbd_id, location_id)
    data = data[data.measure_id == MEASURES["Deaths"]]
    return data


def extract_exposure(
    entity: Union[RiskFactor, AlternativeRiskFactor], location_id: int, **get_draws_kwargs
) -> pd.DataFrame:
    if entity.kind == "risk_factor":
        data = gbd.get_exposure(entity.gbd_id, location_id, **get_draws_kwargs)
        allowable_measures = [
            MEASURES["Proportion"],
            MEASURES["Continuous"],
            MEASURES["Prevalence"],
        ]
        proper_measure_id = set(data.measure_id).intersection(allowable_measures)
        if len(proper_measure_id) != 1:
            raise DataAbnormalError(
                f"Exposure data have {len(proper_measure_id)} measure id(s). Data should have"
                f"exactly one id out of {allowable_measures} but came back with {proper_measure_id}."
            )
        else:
            data = data[data.measure_id == proper_measure_id.pop()]

    else:  # alternative_risk_factor
        data = gbd.get_auxiliary_data("exposure", entity.kind, entity.name, location_id)

    return data


def extract_exposure_standard_deviation(
    entity: Union[RiskFactor, AlternativeRiskFactor], location_id: int
) -> pd.DataFrame:
    if entity.kind == "risk_factor" and entity.name in OTHER_MEID:
        data = gbd.get_modelable_entity_draws(OTHER_MEID[entity.name], location_id)
    elif entity.kind == "risk_factor":
        data = gbd.get_exposure_standard_deviation(entity.gbd_id, location_id)
    else:  # alternative_risk_factor
        data = gbd.get_auxiliary_data(
            "exposure_standard_deviation", entity.kind, entity.name, location_id
        )
    return data


def extract_exposure_distribution_weights(
    entity: Union[RiskFactor, AlternativeRiskFactor], location_id: int
) -> pd.DataFrame:
    data = gbd.get_auxiliary_data(
        "exposure_distribution_weights", entity.kind, entity.name, location_id
    )
    return data


def extract_relative_risk(entity: RiskFactor, location_id: int) -> pd.DataFrame:
    data = gbd.get_relative_risk(entity.gbd_id, location_id)
    data = filter_to_most_detailed_causes(data)
    return data


def extract_population_attributable_fraction(
    entity: Union[RiskFactor, Etiology], location_id: int
) -> pd.DataFrame:
    data = gbd.get_paf(entity.gbd_id, location_id)
    data = data[data.metric_id == METRICS["Percent"]]
    data = data[data.measure_id.isin([MEASURES["YLDs"], MEASURES["YLLs"]])]
    data = filter_to_most_detailed_causes(data)
    return data


def extract_mediation_factors(entity: RiskFactor, location_id: int) -> pd.DataFrame:
    data = gbd.get_auxiliary_data("mediation_factor", entity.kind, entity.name, location_id)
    return data


def extract_estimate(entity: Covariate, location_id: int) -> pd.DataFrame:
    data = gbd.get_covariate_estimate(int(entity.gbd_id), location_id)
    return data


def extract_utilization_rate(entity: HealthcareEntity, location_id: int) -> pd.DataFrame:
    data = gbd.get_modelable_entity_draws(entity.gbd_id, location_id)
    return data


def extract_structure(entity: Population, location_id: int) -> pd.DataFrame:
    data = gbd.get_population(location_id)
    return data


def extract_theoretical_minimum_risk_life_expectancy(
    entity: Population, location_id: int
) -> pd.DataFrame:
    data = gbd.get_theoretical_minimum_risk_life_expectancy()
    return data



from gbd_mapping import causes
from gbd_mapping.base_template import Restrictions
from gbd_mapping.id import rei_id

from .alternative_risk_factor_template import (
    AlternativeRiskFactor,
    AlternativeRiskFactors,
)

alternative_risk_factors = AlternativeRiskFactors(
    child_underweight=AlternativeRiskFactor(
        name="child_underweight",
        kind="alternative_risk_factor",
        gbd_id=rei_id(94),
        level=4,
        most_detailed=True,
        distribution="ensemble",
        population_attributable_fraction_calculation_type="categorical",
        restrictions=Restrictions(
            male_only=False,
            female_only=False,
            yll_only=False,
            yld_only=False,
            yll_age_group_id_start=4,
            yll_age_group_id_end=235,
            yld_age_group_id_start=2,
            yld_age_group_id_end=235,
        ),
        affected_causes=(
            causes.all_causes,
            causes.communicable_maternal_neonatal_and_nutritional_diseases,
            causes.diarrheal_diseases,
            causes.lower_respiratory_infections,
            causes.measles,
            causes.nutritional_deficiencies,
            causes.protein_energy_malnutrition,
            causes.respiratory_infections_and_tuberculosis,
            causes.enteric_infections,
            causes.other_infectious_diseases,
            causes.all_causes,
            causes.communicable_maternal_neonatal_and_nutritional_diseases,
            causes.diarrheal_diseases,
            causes.lower_respiratory_infections,
            causes.measles,
            causes.nutritional_deficiencies,
            causes.protein_energy_malnutrition,
            causes.respiratory_infections_and_tuberculosis,
            causes.enteric_infections,
            causes.other_infectious_diseases,
        ),
        population_attributable_fraction_of_one_causes=(causes.protein_energy_malnutrition,),
    ),
    child_wasting=AlternativeRiskFactor(
        name="child_wasting",
        kind="alternative_risk_factor",
        gbd_id=rei_id(240),
        level=4,
        most_detailed=True,
        distribution="ensemble",
        population_attributable_fraction_calculation_type="categorical",
        restrictions=Restrictions(
            male_only=False,
            female_only=False,
            yll_only=False,
            yld_only=False,
            yll_age_group_id_start=4,
            yll_age_group_id_end=235,
            yld_age_group_id_start=2,
            yld_age_group_id_end=235,
        ),
        affected_causes=(
            causes.all_causes,
            causes.communicable_maternal_neonatal_and_nutritional_diseases,
            causes.diarrheal_diseases,
            causes.lower_respiratory_infections,
            causes.measles,
            causes.nutritional_deficiencies,
            causes.protein_energy_malnutrition,
            causes.respiratory_infections_and_tuberculosis,
            causes.enteric_infections,
            causes.other_infectious_diseases,
            causes.all_causes,
            causes.communicable_maternal_neonatal_and_nutritional_diseases,
            causes.diarrheal_diseases,
            causes.lower_respiratory_infections,
            causes.measles,
            causes.nutritional_deficiencies,
            causes.protein_energy_malnutrition,
            causes.respiratory_infections_and_tuberculosis,
            causes.enteric_infections,
            causes.other_infectious_diseases,
        ),
        population_attributable_fraction_of_one_causes=(causes.protein_energy_malnutrition,),
    ),
    child_stunting=AlternativeRiskFactor(
        name="child_stunting",
        kind="alternative_risk_factor",
        gbd_id=rei_id(241),
        level=4,
        most_detailed=True,
        distribution="ensemble",
        population_attributable_fraction_calculation_type="categorical",
        restrictions=Restrictions(
            male_only=False,
            female_only=False,
            yll_only=False,
            yld_only=False,
            yll_age_group_id_start=4,
            yll_age_group_id_end=5,
            yld_age_group_id_start=4,
            yld_age_group_id_end=5,
        ),
        affected_causes=(
            causes.all_causes,
            causes.communicable_maternal_neonatal_and_nutritional_diseases,
            causes.diarrheal_diseases,
            causes.lower_respiratory_infections,
            causes.measles,
            causes.respiratory_infections_and_tuberculosis,
            causes.enteric_infections,
            causes.other_infectious_diseases,
            causes.all_causes,
            causes.communicable_maternal_neonatal_and_nutritional_diseases,
            causes.diarrheal_diseases,
            causes.lower_respiratory_infections,
            causes.measles,
            causes.respiratory_infections_and_tuberculosis,
            causes.enteric_infections,
            causes.other_infectious_diseases,
        ),
        population_attributable_fraction_of_one_causes=(),
    ),
)



from vivarium_inputs.__about__ import (
    __author__,
    __copyright__,
    __email__,
    __license__,
    __summary__,
    __title__,
    __uri__,
)
from vivarium_inputs._version import __version__
from vivarium_inputs.interface import (
    get_age_bins,
    get_demographic_dimensions,
    get_measure,
    get_population_structure,
    get_raw_data,
    get_theoretical_minimum_risk_life_expectancy,
)



"""Global constants, errors, and module imports for inputs processing."""
from gbd_mapping import ModelableEntity, causes, risk_factors

# The purpose of this import block is to mask the dependency on internal
# IHME data and allow CI and automated testing to work.
try:
    from vivarium_gbd_access import gbd

    try:
        from gbd_artifacts.exceptions import NoBestVersionError
        from get_draws.api import EmptyDataFrameException, InputsException
    except ModuleNotFoundError:
        raise RuntimeError("Problem importing gbd_artifacts.exceptions or get_draws.api.")

except ModuleNotFoundError:

    class GbdDummy:
        """Mock class to wrap internal dependency."""

        def __getattr__(self, item):
            raise ModuleNotFoundError("Required package vivarium_gbd_access not found.")

    gbd = GbdDummy()

    class NoBestVersionError(Exception):
        """Mock class for gbd exception"""

        pass

    class EmptyDataFrameException(Exception):
        """Mock class for gbd exception"""

        pass

    class InputsException(Exception):
        """Mock class for gbd exception"""

        pass


class VivariumInputsError(Exception):
    """Base exception for errors in vivarium_inputs."""

    pass


class DataDoesNotExistError(VivariumInputsError, FileNotFoundError):
    """Exception raised when requested GBD data does not exist."""

    pass


class DataAbnormalError(VivariumInputsError, ValueError):
    """Exception raised when raw GBD data does not have an expected format."""

    pass


class DataTransformationError(VivariumInputsError, ValueError):
    """Exception raised when data is incorrectly transformed."""

    pass


class InvalidQueryError(VivariumInputsError, ValueError):
    """Exception raised when the user makes an invalid query."""

    pass


METRICS = {
    # Mapping of GBD metric names to their GBD ids.
    "Number": 1,
    "Percent": 2,
    "Rate": 3,
    "Rank": 4,
    "Years": 5,
    "p-value": 6,
    "MDG p-value": 7,
    "Probability of death": 8,
    "Index score": 9,
}

MEASURES = {
    # Mapping of GBD measure names to their GBD ids.
    "Deaths": 1,
    "DALYs": 2,
    "YLDs": 3,
    "YLLs": 4,
    "Prevalence": 5,
    "Incidence rate": 6,
    "Remission rate": 7,
    "Duration": 8,
    "Excess mortality rate": 9,
    "Prevalence * excess mortality rate": 10,
    "Relative risk": 11,
    "Standardized mortality ratio": 12,
    "With-condition mortality rate": 13,
    "All-cause mortality rate": 14,
    "Cause-specific mortality rate": 15,
    "Other cause mortality rate": 16,
    "Case fatality rate": 17,
    "Proportion": 18,
    "Continuous": 19,
    "Survival Rate": 20,
    "Disability Weight": 21,
    "Chronic Prevalence": 22,
    "Acute Prevalence": 23,
    "Acute Incidence": 24,
    "Maternal mortality ratio": 25,
    "Life expectancy": 26,
    "Probability of death": 27,
    "HALE (Healthy life expectancy)": 28,
    "Summary exposure value": 29,
    "Life expectancy no-shock hiv free": 30,
    "Life expectancy no-shock with hiv": 31,
    "Probability of death no-shock hiv free": 32,
    "Probability of death no-shock with hiv": 33,
    "Mortality risk": 34,
    "Short term prevalence": 35,
    "Long term prevalence": 36,
    "Life expectancy decomposition by cause": 37,
    "Birth prevalence": 38,
    "Susceptible population fraction": 39,
    "With Condition population fraction": 40,
    "Susceptible incidence": 41,
    "Total incidence": 42,
    "HAQ Index (Healthcare Access and Quality Index)": 43,
    "Population": 44,
    "Fertility": 45,
}

# List of standard distribution columns
DISTRIBUTION_COLUMNS = [
    "exp",
    "gamma",
    "invgamma",
    "llogis",
    "gumbel",
    "invweibull",
    "weibull",
    "lnorm",
    "norm",
    "glnorm",
    "betasr",
    "mgamma",
    "mgumbel",
]
# List of standard covariate values
COVARIATE_VALUE_COLUMNS = ["mean_value", "upper_value", "lower_value"]
# List of standard demographic id column names.
DEMOGRAPHIC_COLUMNS = ["location_id", "sex_id", "age_group_id", "year_id"]
# List of standard GBD draw column names.
DRAW_COLUMNS = [f"draw_{i}" for i in range(1000)]
# Mapping of GBD sex ids
SEXES = {"Male": 1, "Female": 2, "Combined": 3}
# Mapping of non-standard age group ids sometimes found in GBD data
SPECIAL_AGES = {"all_ages": 22, "age_standardized": 27}

# Cause-risk pair where risk may have a protective effect on a certain cause with negative paf
PROTECTIVE_CAUSE_RISK_PAIRS = {
    "high_body_mass_index_in_adults": [
        causes.neoplasms,
        causes.breast_cancer,
        causes.esophageal_cancer,
    ]
}

# Keep track of special cases for the sim validator boundary checks where the
# standard boundary in sim validators won't cut it
BOUNDARY_SPECIAL_CASES = {
    "excess_mortality_rate": {
        "Ecuador": {"measles": 250_000_000},
        "China": {"subarachnoid_hemorrhage": 3_775},
        "Russian Federation": {"subarachnoid_hemorrhage": 800},
        "South Africa": {"subarachnoid_hemorrhage": 500},
        "Bangladesh": {"subarachnoid_hemorrhage": 2_750},
        "Ukraine": {"subarachnoid_hemorrhage": 500},
    }
}

RISKS_WITH_NEGATIVE_PAF = [
    risk_factors.vitamin_a_deficiency.name,
    risk_factors.child_wasting.name,
    risk_factors.high_body_mass_index_in_adults.name,
]

# residual cat is added by get_draws but all cats modeled for lbwsg so
# has to be removed
EXTRA_RESIDUAL_CATEGORY = {risk_factors.low_birth_weight_and_short_gestation.name: "cat125"}

# Some data contain very small (5.56e-311) values that produce floating point errors, clip them
MINIMUM_EXPOSURE_VALUE = 1e-10

PROBLEMATIC_RISKS = {
    risk_factors.zinc_deficiency.name: "zinc deficiency relative risk data breaks central comp interpolation."
}

NON_MAX_TMREL = {risk_factors.low_birth_weight_and_short_gestation.name: "cat56"}

# LBWSG paf has data outside neonatal preterm birth age restrictions (but is all 1.0) - K.W. 4/2/19
PAF_OUTSIDE_AGE_RESTRICTIONS = {
    risk_factors.low_birth_weight_and_short_gestation.name: [causes.neonatal_preterm_birth]
}


def EXCLUDE_ABNORMAL_DATA(entity, context):
    # Add expressions to the list that evaluate to True if the subsequent validation
    #  check should be excluded
    ABNORMAL_CONDITIONS = [
        # The data for sex is incorrect in this situation (location is Pakistan)
        (
            entity.kind == "sequela"
            and entity.name == "moderate_wasting_with_edema"
            and context.context_data["location_id"] == 165
        ),
    ]
    for condition in ABNORMAL_CONDITIONS:
        if condition:
            return True
    return False


# GBD uses modelable entity 10488, which is attached to anemia not iron deficiency, for iron deficiency SD
OTHER_MEID = {risk_factors.iron_deficiency.name: 10488}


class Population(ModelableEntity):
    """Entity wrapper for querying population measures."""

    def __init__(self):
        super().__init__(name="population", kind="population", gbd_id=None)



import numpy as np
import pandas as pd
import pytest

from vivarium_inputs.globals import DataTransformationError
from vivarium_inputs.validation import sim


@pytest.fixture
def mock_validation_context():
    years = pd.DataFrame({"year_start": range(1990, 2017), "year_end": range(1991, 2018)})
    age_bins = pd.DataFrame(
        {
            "age_group_id": [1, 2, 3, 4, 5],
            "age_group_name": ["youngest", "young", "middle", "older", "oldest"],
            "age_start": [0, 1, 15, 45, 60],
            "age_end": [1, 15, 45, 60, 100],
        }
    )
    context = sim.SimulationValidationContext(
        location="United States", years=years, age_bins=age_bins
    )

    return context


@pytest.mark.parametrize("location", ("Kenya", "Papua New Guinea"))
def test__validate_location_column_pass(mock_validation_context, location):
    mock_validation_context["location"] = location
    df = pd.DataFrame({"location": [location], "other": [0]}).set_index(["location", "other"])
    sim.validate_location_column(df, mock_validation_context)


@pytest.mark.parametrize(
    "locations,expected_location",
    ((["Kenya", "Kenya"], "Egypt"), (["Algeria", "Nigeria"], "Algeria")),
    ids=("mismatch", "multiple"),
)
def test_validate_location_column_fail(mock_validation_context, locations, expected_location):
    mock_validation_context["location"] = expected_location
    df = pd.DataFrame({"location": locations, "other": [0] * len(locations)}).set_index(
        ["location", "other"]
    )
    with pytest.raises(DataTransformationError):
        sim.validate_location_column(df, mock_validation_context)


def test_validate_sex_column_pass():
    df = pd.DataFrame({"sex": ["Male", "Female"], "other": [0] * 2}).set_index(
        ["sex", "other"]
    )
    sim.validate_sex_column(df)


@pytest.mark.parametrize(
    "sexes",
    (("Male", "female"), ("Male", "Male"), ("Female", "Female")),
    ids=("lowercase", "missing_female", "missing_male"),
)
def test_validate_sex_column_fail(sexes):
    df = pd.DataFrame({"sex": sexes, "other": [0] * len(sexes)}).set_index(["sex", "other"])
    with pytest.raises(DataTransformationError):
        sim.validate_sex_column(df)


def test_validate_age_columns_pass(mock_validation_context):
    ages = mock_validation_context["age_bins"].filter(["age_start", "age_end"])
    # Shuffle the rows and set index
    ages = ages.sample(frac=1).reset_index(drop=True)
    ages = ages.set_index(
        pd.IntervalIndex.from_arrays(ages.age_start, ages.age_end, closed="left", name="age"),
        append=True,
    )
    sim.validate_age_column(ages, mock_validation_context)


def test_validate_age_columns_invalid_age(mock_validation_context):
    ages = mock_validation_context["age_bins"].filter(["age_start", "age_end"]).copy()
    ages.loc[2, "age_start"] = -1
    ages = ages.set_index(
        pd.IntervalIndex.from_arrays(ages.age_start, ages.age_end, closed="left", name="age"),
        append=True,
    )
    with pytest.raises(DataTransformationError):
        sim.validate_age_column(ages, mock_validation_context)


def test_validate_age_columns_missing_group(mock_validation_context):
    ages = mock_validation_context["age_bins"].filter(["age_start", "age_end"])
    ages = ages.drop(2)
    ages = ages.set_index(
        pd.IntervalIndex.from_arrays(ages.age_start, ages.age_end, closed="left", name="age"),
        append=True,
    )
    with pytest.raises(DataTransformationError):
        sim.validate_age_column(ages, mock_validation_context)


def test_validate_year_columns_pass(mock_validation_context):
    years = mock_validation_context["years"]
    # Shuffle the rows and set index
    years = years.sample(frac=1).reset_index(drop=True)
    years = years.set_index(
        pd.IntervalIndex.from_arrays(
            years.year_start, years.year_end, closed="left", name="year"
        ),
        append=True,
    )
    sim.validate_year_column(years, mock_validation_context)


def test_validate_year_columns_invalid_year(mock_validation_context):
    years = mock_validation_context["years"].copy()
    years.loc[2, "year_end"] = 20100
    # Shuffle the rows and set index
    years = years.sample(frac=1).reset_index(drop=True)
    years = years.set_index(
        pd.IntervalIndex.from_arrays(
            years.year_start, years.year_end, closed="left", name="year"
        ),
        append=True,
    )
    with pytest.raises(DataTransformationError):
        sim.validate_year_column(years, mock_validation_context)


def test_validate_year_columns_missing_group(mock_validation_context):
    years = mock_validation_context["years"].sort_values(["year_start", "year_end"])
    years = years.drop(0)
    years = years.set_index(
        pd.IntervalIndex.from_arrays(
            years.year_start, years.year_end, closed="left", name="year"
        ),
        append=True,
    )
    with pytest.raises(DataTransformationError):
        sim.validate_year_column(years, mock_validation_context)


@pytest.mark.parametrize("values", [(-1, 2, 3)], ids=["integers"])
def test_validate_value_column_pass(values):
    df = pd.DataFrame({"value": values})
    sim.validate_value_column(df)


@pytest.mark.parametrize(
    "values", [(1, 2, np.inf), (1, np.nan, 2)], ids=["infinity", "missing"]
)
def test_validate_value_column_fail(values):
    df = pd.DataFrame({"value": values})
    with pytest.raises(DataTransformationError):
        sim.validate_value_column(df)


@pytest.mark.parametrize(
    "values, ids, restriction_type, fill",
    [
        ((1, 1, 1, 1, 1), (1, 5), "outer", 0.0),
        ((0, 0, 1, 1, 1), (3, 5), "outer", 0.0),
        ((0, 1, 1, 1, 0), (2, 4), "outer", 0.0),
        ((1, 1, 1, 0, 0), (1, 3), "outer", 0.0),
        ((2, 2, 2, 1, 1), (1, 3), "outer", 1.0),
    ],
    ids=("no_restr", "left_restr", "outer_restr", "right_restr", "nonzero_fill"),
)
def test_check_age_restrictions(
    mocker, mock_validation_context, values, ids, restriction_type, fill
):
    entity = mocker.patch(
        "vivarium_inputs.validation.sim.utilities.get_age_group_ids_by_restriction"
    )
    entity.return_value = ids
    age_bins = mock_validation_context["age_bins"]
    idx = pd.IntervalIndex.from_arrays(
        age_bins.age_start, age_bins.age_end, closed="left", name="age"
    )
    df = pd.DataFrame({"value": values}, index=idx)
    sim.check_age_restrictions(df, entity, restriction_type, fill, mock_validation_context)


@pytest.mark.parametrize(
    "values, ids, restriction_type, fill",
    [
        ((1, 1, 1, 1, 1), (1, 4), "outer", 0.0),
        ((0, 1, 1, 1, 1), (1, 3), "outer", 0.0),
        ((1, 1, 1, 1, 0), (1, 3), "outer", 0.0),
        ((2, 2, 2, 2, 1), (2, 5), "outer", 1.0),
    ],
    ids=("both_sides", "left_side", "right_side", "nonzero_fill"),
)
def test_check_age_restrictions_fail(
    mocker, mock_validation_context, values, ids, restriction_type, fill
):
    entity = mocker.patch(
        "vivarium_inputs.validation.sim.utilities.get_age_group_ids_by_restriction"
    )
    entity.return_value = ids
    age_bins = mock_validation_context["age_bins"]
    idx = pd.IntervalIndex.from_arrays(
        age_bins.age_start, age_bins.age_end, closed="left", name="age"
    )
    df = pd.DataFrame({"value": values}, index=idx)
    with pytest.raises(DataTransformationError):
        sim.check_age_restrictions(
            df, entity, restriction_type, fill, mock_validation_context
        )


@pytest.mark.parametrize(
    "values, restrictions, fill",
    [
        ((1, 1, 1, 1), (False, False), 0.0),
        ((1, 1, 0, 0), (True, False), 0.0),
        ((0, 0, 1, 1), (False, True), 0.0),
        ((1, 1, 2, 2), (False, True), 1.0),
    ],
    ids=("None", "male", "female", "nonzero_fill"),
)
def test_check_sex_restrictions(values, restrictions, fill):
    df = pd.DataFrame(
        {"sex": ["Male", "Male", "Female", "Female"], "other": [0] * 4, "value": values}
    )
    df = df.set_index(["sex", "other"])
    sim.check_sex_restrictions(df, restrictions[0], restrictions[1], fill)


@pytest.mark.parametrize(
    "values, restrictions, fill",
    [
        ((1, 1, 1, 0), (True, False), 0.0),
        ((0, 1, 1, 1), (False, True), 0.0),
        ((1, 2, 2, 2), (False, True), 1.0),
    ],
    ids=("male", "female", "nonzero_fill"),
)
def test_check_sex_restrictions_fail(values, restrictions, fill):
    df = pd.DataFrame(
        {"sex": ["Male", "Male", "Female", "Female"], "other": [0] * 4, "value": values}
    )
    df = df.set_index(["sex", "other"])
    with pytest.raises(DataTransformationError):
        sim.check_sex_restrictions(df, restrictions[0], restrictions[1], fill)


@pytest.mark.parametrize(
    "df",
    [
        pd.DataFrame(columns=["a", "b", "c", "d"]),
        pd.DataFrame({"a": [0], "b": [0], "c": [0], "d": [0]}),
        pd.DataFrame({"a": [1] * 3, "b": [2] * 3, "c": [3] * 3, "d": [np.NaN] * 3}),
    ],
)
def test_validate_expected_index_columns_pass(df):
    cols = ["a", "b", "c", "d"]
    sim.validate_expected_index_and_columns([None], df.index.names, cols, df.columns)
    df = df.set_index(cols[:2])
    sim.validate_expected_index_and_columns(cols[:2], df.index.names, cols[2:], df.columns)
    df = df.set_index(cols[2:], append=True)
    sim.validate_expected_index_and_columns(cols, df.index.names, [], df.columns)


@pytest.mark.parametrize(
    "df",
    [
        pd.DataFrame(columns=["a", "b", "c", "d"]),
        pd.DataFrame({"a": [0], "b": [0], "c": [0], "d": [0]}),
        pd.DataFrame({"a": [1] * 3, "b": [2] * 3, "c": [3] * 3, "d": [np.NaN] * 3}),
    ],
)
def test_validate_expected_columns_fail(df):
    cols = ["a", "b", "c", "d"]
    with pytest.raises(DataTransformationError, match="missing columns"):
        sim.validate_expected_index_and_columns(
            [None], df.index.names, cols + ["e"], df.columns
        )
    with pytest.raises(DataTransformationError, match="missing index names"):
        data = df.set_index(cols[:2])
        sim.validate_expected_index_and_columns(
            cols[:3], data.index.names, cols[:2], data.columns
        )

    with pytest.raises(DataTransformationError, match="extra columns"):
        sim.validate_expected_index_and_columns([None], df.index.names, cols[1:], df.columns)
    with pytest.raises(DataTransformationError, match="extra index names"):
        data = df.set_index(cols[:2])
        sim.validate_expected_index_and_columns(
            cols[:1], data.index.names, cols[:2], data.columns
        )



import pandas as pd
import pytest

from vivarium_inputs.globals import DataAbnormalError, DataTransformationError
from vivarium_inputs.validation import shared

test_data = [
    (
        pd.DataFrame({"a": [0, 1], "b": [2, 20]}),
        1,
        "lower",
        ["a", "b"],
        True,
        "values below",
        DataAbnormalError,
    ),
    (
        pd.DataFrame({"a": [0, 1], "b": [2, 20]}),
        0,
        "lower",
        ["a", "b"],
        False,
        "values below or equal to",
        ValueError,
    ),
    (
        pd.DataFrame({"a": [0], "b": [10], "c": [100]}),
        10,
        "lower",
        ["a"],
        True,
        "values below",
        ValueError,
    ),
    (
        pd.DataFrame({"a": [0, 1], "b": [2, 20]}),
        20,
        "upper",
        ["a", "b"],
        False,
        "values above",
        DataAbnormalError,
    ),
    (
        pd.DataFrame({"a": [0, 1], "b": [2, 20]}, index=[0, 1]),
        pd.Series([5, 10], index=[0, 1]),
        "upper",
        ["a", "b"],
        True,
        "above the expected",
        DataTransformationError,
    ),
]


@pytest.mark.parametrize(
    "data, boundary, boundary_type, val_cols, inclusive, match, error", test_data
)
def test_check_value_columns_boundary_fail_warn(
    caplog, data, boundary, boundary_type, val_cols, inclusive, match, error
):
    with pytest.raises(error, match=match):
        shared.check_value_columns_boundary(
            data, boundary, boundary_type, val_cols, inclusive=inclusive, error=error
        )

    shared.check_value_columns_boundary(
        data, boundary, boundary_type, val_cols, inclusive=inclusive, error=None
    )
    assert match in caplog.text


test_data = [
    (pd.DataFrame({"a": [0, 1], "b": [2, 20]}), 0, "lower", ["a", "b"], True),
    (pd.DataFrame({"a": [0, 1], "b": [2, 20]}), -1, "lower", ["a", "b"], False),
    (pd.DataFrame({"a": [0], "b": [10], "c": [100]}), 10, "lower", ["b", "c"], True),
    (pd.DataFrame({"a": [0, 1], "b": [2, 20]}), 20, "upper", ["a", "b"], True),
    (
        pd.DataFrame({"a": [0, 1], "b": [2, 20]}, index=[0, 1]),
        pd.Series([0, 1], index=[0, 1]),
        "upper",
        ["a"],
        True,
    ),
]


@pytest.mark.parametrize("data, boundary, boundary_type, val_cols, inclusive", test_data)
def test_check_value_columns_boundary_pass(
    data, boundary, boundary_type, val_cols, inclusive
):
    shared.check_value_columns_boundary(
        data, boundary, boundary_type, val_cols, inclusive=inclusive, error=ValueError
    )



"""Access to vivarium simulation input data."""
from typing import Union

import pandas as pd
from gbd_mapping import ModelableEntity

import vivarium_inputs.validation.sim as validation
from vivarium_inputs import core, extract, utilities, utility_data
from vivarium_inputs.globals import Population


def get_measure(
    entity: ModelableEntity, measure: str, location: str, **get_draws_kwargs
) -> pd.DataFrame:
    """Pull GBD data for measure and entity and prep for simulation input,
    including scrubbing all GBD conventions to replace IDs with meaningful
    values or ranges and expanding over all demographic dimensions. To pull data
    using this function, please have at least 50GB of memory available.

    Available measures:

        For entity kind 'sequela':
            incidence_rate, prevalence, birth_prevalence, disability_weight

        For entity kind 'cause':
            incidence_rate, prevalence, birth_prevalence, disability_weight,
            remission_rate, cause_specific_mortality_rate, excess_mortality_rate

        For entity kind 'risk_factor':
            exposure, exposure_standard_deviation, exposure_distribution_weights,
            relative_risk, population_attributable_fraction, mediation_factors

        For entity kind 'etiology':
            population_attributable_fraction

        For entity kind 'alternative_risk_factor':
            exposure, exposure_standard_deviation, exposure_distribution_weights

        For entity kind 'covariate':
            estimate

    Parameters
    ----------
    entity
        Entity for which to pull `measure`.
    measure
        Measure for which to pull data, should be a measure available for the
        kind of entity which `entity` is.
    location
        Location for which to pull data.

    Returns
    -------
    pandas.DataFrame
        Dataframe standardized to the format expected by `vivarium` simulations.

    """
    data = core.get_data(entity, measure, location, **get_draws_kwargs)
    data = utilities.scrub_gbd_conventions(data, location)
    validation.validate_for_simulation(data, entity, measure, location)
    data = utilities.split_interval(data, interval_column="age", split_column_prefix="age")
    data = utilities.split_interval(data, interval_column="year", split_column_prefix="year")
    return utilities.sort_hierarchical_data(data)


def get_population_structure(location: str) -> pd.DataFrame:
    """Pull GBD population data for the given location and standardize to the
    expected simulation input format, including scrubbing all GBD conventions
    to replace IDs with meaningful values or ranges and expanding over all
    demographic dimensions.

    Parameters
    ----------
    location
        Location for which to pull population data.

    Returns
    -------
    pandas.DataFrame
        Dataframe of population data for `location`, standardized to the format
        expected by `vivarium` simulations.

    """
    pop = Population()
    data = core.get_data(pop, "structure", location)
    data = utilities.scrub_gbd_conventions(data, location)
    validation.validate_for_simulation(data, pop, "structure", location)
    data = utilities.split_interval(data, interval_column="age", split_column_prefix="age")
    data = utilities.split_interval(data, interval_column="year", split_column_prefix="year")
    return utilities.sort_hierarchical_data(data)


def get_theoretical_minimum_risk_life_expectancy() -> pd.DataFrame:
    """Pull GBD theoretical minimum risk life expectancy data and standardize
    to the expected simulation input format, including binning age parameters
    as expected by simulations.

    Returns
    -------
    pandas.DataFrame
        Dataframe of theoretical minimum risk life expectancy data, standardized
        to the format expected by `vivarium` simulations with binned age parameters.

    """
    pop = Population()
    data = core.get_data(pop, "theoretical_minimum_risk_life_expectancy", "Global")
    data = utilities.set_age_interval(data)
    validation.validate_for_simulation(
        data, pop, "theoretical_minimum_risk_life_expectancy", "Global"
    )
    data = utilities.split_interval(data, interval_column="age", split_column_prefix="age")
    data = utilities.split_interval(data, interval_column="year", split_column_prefix="year")
    return utilities.sort_hierarchical_data(data)


def get_age_bins() -> pd.DataFrame:
    """Pull GBD age bin data and standardize to the expected simulation input
    format.

    Returns
    -------
    pandas.DataFrame
        Dataframe of age bin data, with bin start and end values as well as bin
        names.

    """
    pop = Population()
    data = core.get_data(pop, "age_bins", "Global")
    data = utilities.set_age_interval(data)
    validation.validate_for_simulation(data, pop, "age_bins", "Global")
    data = utilities.split_interval(data, interval_column="age", split_column_prefix="age")
    data = utilities.split_interval(data, interval_column="year", split_column_prefix="year")
    return utilities.sort_hierarchical_data(data)


def get_demographic_dimensions(location: str) -> pd.DataFrame:
    """Pull the full demographic dimensions for GBD data, standardized to the
    expected simulation input format, including scrubbing all GBD conventions
    to replace IDs with with meaningful values or ranges.

    Parameters
    ----------
    location
        Location for which to pull demographic dimension data.

    Returns
    -------
    pandas.DataFrame
        Dataframe with age and year bins from GBD, sexes, and the given location.

    """
    pop = Population()
    data = core.get_data(pop, "demographic_dimensions", location)
    data = utilities.scrub_gbd_conventions(data, location)
    validation.validate_for_simulation(data, pop, "demographic_dimensions", location)
    data = utilities.split_interval(data, interval_column="age", split_column_prefix="age")
    data = utilities.split_interval(data, interval_column="year", split_column_prefix="year")
    return utilities.sort_hierarchical_data(data)


def get_raw_data(
    entity: ModelableEntity, measure: str, location: str
) -> Union[pd.Series, pd.DataFrame]:
    """Pull raw data from GBD for the requested entity, measure, and location.
    Skip standard raw validation checks in order to return data that can be
    investigated for oddities. The only filter that occurs is by applicable
    measure id, metric id, or to most detailed causes where relevant.

    Available measures:

        For entity kind 'sequela':
            incidence_rate, prevalence, birth_prevalence, disability_weight

        For entity kind 'cause':
            incidence_rate, prevalence, birth_prevalence, disability_weight,
            remission_rate, deaths

        For entity kind 'risk_factor':
            exposure, exposure_standard_deviation, exposure_distribution_weights,
            relative_risk, population_attributable_fraction, mediation_factors

        For entity kind 'etiology':
            population_attributable_fraction

        For entity kind 'alternative_risk_factor':
            exposure, exposure_standard_deviation, exposure_distribution_weights

        For entity kind 'covariate':
            estimate

        For entity kind 'population':
            structure, theoretical_minimum_risk_life_expectancy

    Parameters
    ----------
    entity
        Entity for which to extract data.
    measure
        Measure for which to extract data.
    location
        Location for which to extract data.

    Returns
    -------
    Union[pandas.Series, pandas.DataFrame]
        Data for the entity-measure pair and specific location requested, with no
        formatting or reshaping.
    """
    location_id = utility_data.get_location_id(location)
    data = extract.extract_data(entity, measure, location_id, validate=False)
    return data



from itertools import product
from typing import Union

import numpy as np
import pandas as pd
from gbd_mapping import Cause, Covariate, Etiology, RiskFactor, Sequela, causes
from loguru import logger

from vivarium_inputs import extract, utilities, utility_data
from vivarium_inputs.globals import (
    COVARIATE_VALUE_COLUMNS,
    DEMOGRAPHIC_COLUMNS,
    DISTRIBUTION_COLUMNS,
    DRAW_COLUMNS,
    EXTRA_RESIDUAL_CATEGORY,
    MEASURES,
    MINIMUM_EXPOSURE_VALUE,
    DataDoesNotExistError,
    InvalidQueryError,
    Population,
)
from vivarium_inputs.mapping_extension import AlternativeRiskFactor, HealthcareEntity


def get_data(entity, measure: str, location: Union[str, int], **get_draws_kwargs):
    measure_handlers = {
        # Cause-like measures
        "incidence_rate": (get_incidence_rate, ("cause", "sequela")),
        "raw_incidence_rate": (get_raw_incidence_rate, ("cause", "sequela")),
        "prevalence": (get_prevalence, ("cause", "sequela")),
        "birth_prevalence": (get_birth_prevalence, ("cause", "sequela")),
        "disability_weight": (get_disability_weight, ("cause", "sequela")),
        "remission_rate": (get_remission_rate, ("cause",)),
        "cause_specific_mortality_rate": (get_cause_specific_mortality_rate, ("cause",)),
        "excess_mortality_rate": (get_excess_mortality_rate, ("cause",)),
        "deaths": (get_deaths, ("cause",)),
        # Risk-like measures
        "exposure": (
            get_exposure,
            (
                "risk_factor",
                "alternative_risk_factor",
            ),
        ),
        "exposure_standard_deviation": (
            get_exposure_standard_deviation,
            ("risk_factor", "alternative_risk_factor"),
        ),
        "exposure_distribution_weights": (
            get_exposure_distribution_weights,
            ("risk_factor", "alternative_risk_factor"),
        ),
        "relative_risk": (get_relative_risk, ("risk_factor",)),
        "population_attributable_fraction": (
            get_population_attributable_fraction,
            ("risk_factor", "etiology"),
        ),
        # Covariate measures
        "estimate": (get_estimate, ("covariate",)),
        # Population measures
        "structure": (get_structure, ("population",)),
        "theoretical_minimum_risk_life_expectancy": (
            get_theoretical_minimum_risk_life_expectancy,
            ("population",),
        ),
        "age_bins": (get_age_bins, ("population",)),
        "demographic_dimensions": (get_demographic_dimensions, ("population",)),
    }

    if measure not in measure_handlers:
        raise InvalidQueryError(f"No functions available to pull data for measure {measure}.")

    handler, entity_types = measure_handlers[measure]

    if entity.kind not in entity_types:
        raise InvalidQueryError(f"{measure.capitalize()} not available for {entity.kind}.")

    location_id = (
        utility_data.get_location_id(location) if isinstance(location, str) else location
    )
    data = handler(entity, location_id, **get_draws_kwargs)

    if measure in [
        "structure",
        "theoretical_minimum_risk_life_expectancy",
        "estimate",
        "exposure_distribution_weights",
    ]:
        value_cols = ["value"]
    else:
        value_cols = DRAW_COLUMNS

    data = utilities.reshape(data, value_cols=value_cols)

    return data


def get_raw_incidence_rate(entity: Union[Cause, Sequela], location_id: int) -> pd.DataFrame:
    data = extract.extract_data(entity, "incidence_rate", location_id)
    if entity.kind == "cause":
        restrictions_entity = entity
    else:  # sequela
        cause = [c for c in causes if c.sequelae and entity in c.sequelae][0]
        restrictions_entity = cause

    data = utilities.filter_data_by_restrictions(
        data, restrictions_entity, "yld", utility_data.get_age_group_ids()
    )
    data = utilities.normalize(data, fill_value=0)
    data = data.filter(DEMOGRAPHIC_COLUMNS + DRAW_COLUMNS)
    return data


def get_incidence_rate(entity: Union[Cause, Sequela], location_id: int) -> pd.DataFrame:
    data = get_data(entity, "raw_incidence_rate", location_id)
    prevalence = get_data(entity, "prevalence", location_id)
    # Convert from "True incidence" to the incidence rate among susceptibles
    data /= 1 - prevalence
    return data.fillna(0)


def get_prevalence(entity: Union[Cause, Sequela], location_id: int) -> pd.DataFrame:
    data = extract.extract_data(entity, "prevalence", location_id)
    if entity.kind == "cause":
        restrictions_entity = entity
    else:  # sequela
        cause = [c for c in causes if c.sequelae and entity in c.sequelae][0]
        restrictions_entity = cause

    data = utilities.filter_data_by_restrictions(
        data, restrictions_entity, "yld", utility_data.get_age_group_ids()
    )
    data = utilities.normalize(data, fill_value=0)
    data = data.filter(DEMOGRAPHIC_COLUMNS + DRAW_COLUMNS)
    return data


def get_birth_prevalence(entity: Union[Cause, Sequela], location_id: int) -> pd.DataFrame:
    data = extract.extract_data(entity, "birth_prevalence", location_id)
    data = data.filter(["year_id", "sex_id", "location_id"] + DRAW_COLUMNS)
    data = utilities.normalize(data, fill_value=0)
    return data


def get_disability_weight(entity: Union[Cause, Sequela], location_id: int) -> pd.DataFrame:
    if entity.kind == "cause":
        data = utility_data.get_demographic_dimensions(location_id, draws=True, value=0.0)
        data = data.set_index(
            utilities.get_ordered_index_cols(data.columns.difference(DRAW_COLUMNS))
        )
        if entity.sequelae:
            for sequela in entity.sequelae:
                try:
                    prevalence = get_data(sequela, "prevalence", location_id)
                except DataDoesNotExistError:
                    # sequela prevalence does not exist so no point continuing with this sequela
                    continue
                disability = get_data(sequela, "disability_weight", location_id)
                disability.index = disability.index.set_levels(
                    [location_id], level="location_id"
                )
                data += prevalence * disability
        cause_prevalence = get_data(entity, "prevalence", location_id)
        data = (data / cause_prevalence).fillna(0).reset_index()
    else:  # entity.kind == 'sequela'
        try:
            data = extract.extract_data(entity, "disability_weight", location_id)
            data = utilities.normalize(data)
            cause = [c for c in causes if c.sequelae and entity in c.sequelae][0]
            data = utilities.clear_disability_weight_outside_restrictions(
                data, cause, 0.0, utility_data.get_age_group_ids()
            )
            data = data.filter(DEMOGRAPHIC_COLUMNS + DRAW_COLUMNS)
        except (IndexError, DataDoesNotExistError):
            logger.warning(
                f"{entity.name.capitalize()} has no disability weight data. All values will be 0."
            )
            data = utility_data.get_demographic_dimensions(location_id, draws=True, value=0.0)
    return data


def get_remission_rate(entity: Cause, location_id: int) -> pd.DataFrame:
    data = extract.extract_data(entity, "remission_rate", location_id)

    data = utilities.filter_data_by_restrictions(
        data, entity, "yld", utility_data.get_age_group_ids()
    )
    data = utilities.normalize(data, fill_value=0)
    data = data.filter(DEMOGRAPHIC_COLUMNS + DRAW_COLUMNS)
    return data


def get_cause_specific_mortality_rate(entity: Cause, location_id: int) -> pd.DataFrame:
    deaths = get_data(entity, "deaths", location_id)  # population isn't by draws
    pop = get_data(Population(), "structure", location_id)
    data = deaths.join(pop, lsuffix="_deaths", rsuffix="_pop")
    data[DRAW_COLUMNS] = data[DRAW_COLUMNS].divide(data.value, axis=0)
    return data.drop(["value"], axis="columns")


def get_excess_mortality_rate(entity: Cause, location_id: int) -> pd.DataFrame:
    csmr = get_data(entity, "cause_specific_mortality_rate", location_id)
    prevalence = get_data(entity, "prevalence", location_id)
    data = (csmr / prevalence).fillna(0)
    data = data.replace([np.inf, -np.inf], 0)
    return data


def get_deaths(entity: Cause, location_id: int) -> pd.DataFrame:
    data = extract.extract_data(entity, "deaths", location_id)
    data = utilities.filter_data_by_restrictions(
        data, entity, "yll", utility_data.get_age_group_ids()
    )
    data = utilities.normalize(data, fill_value=0)
    data = data.filter(DEMOGRAPHIC_COLUMNS + DRAW_COLUMNS)
    return data


def get_exposure(
    entity: Union[RiskFactor, AlternativeRiskFactor], location_id: int, **get_draws_kwargs
) -> pd.DataFrame:
    data = extract.extract_data(entity, "exposure", location_id, **get_draws_kwargs)
    data = data.drop("modelable_entity_id", "columns")

    if entity.name in EXTRA_RESIDUAL_CATEGORY:
        cat = EXTRA_RESIDUAL_CATEGORY[entity.name]
        data = data.drop(labels=data.query("parameter == @cat").index)
        data[DRAW_COLUMNS] = data[DRAW_COLUMNS].clip(lower=MINIMUM_EXPOSURE_VALUE)

    if entity.kind in ["risk_factor", "alternative_risk_factor"]:
        data = utilities.filter_data_by_restrictions(
            data, entity, "outer", utility_data.get_age_group_ids()
        )

    if entity.distribution in ["dichotomous", "ordered_polytomous", "unordered_polytomous"]:
        tmrel_cat = utility_data.get_tmrel_category(entity)
        exposed = data[data.parameter != tmrel_cat]
        unexposed = data[data.parameter == tmrel_cat]

        #  FIXME: We fill 1 as exposure of tmrel category, which is not correct.
        data = pd.concat(
            [
                utilities.normalize(exposed, fill_value=0),
                utilities.normalize(unexposed, fill_value=1),
            ],
            ignore_index=True,
        )

        # normalize so all categories sum to 1
        cols = list(set(data.columns).difference(DRAW_COLUMNS + ["parameter"]))
        sums = data.groupby(cols)[DRAW_COLUMNS].sum()
        data = (
            data.groupby("parameter")
            .apply(lambda df: df.set_index(cols).loc[:, DRAW_COLUMNS].divide(sums))
            .reset_index()
        )
    else:
        data = utilities.normalize(data, fill_value=0)
    data = data.filter(DEMOGRAPHIC_COLUMNS + DRAW_COLUMNS + ["parameter"])
    return data


def get_exposure_standard_deviation(
    entity: Union[RiskFactor, AlternativeRiskFactor], location_id: int
) -> pd.DataFrame:
    data = extract.extract_data(entity, "exposure_standard_deviation", location_id)
    data = data.drop("modelable_entity_id", "columns")

    exposure = extract.extract_data(entity, "exposure", location_id)
    valid_age_groups = utilities.get_exposure_and_restriction_ages(exposure, entity)
    data = data[data.age_group_id.isin(valid_age_groups)]

    data = utilities.normalize(data, fill_value=0)
    data = data.filter(DEMOGRAPHIC_COLUMNS + DRAW_COLUMNS)
    return data


def get_exposure_distribution_weights(
    entity: Union[RiskFactor, AlternativeRiskFactor], location_id: int
) -> pd.DataFrame:
    data = extract.extract_data(entity, "exposure_distribution_weights", location_id)

    exposure = extract.extract_data(entity, "exposure", location_id)
    valid_ages = utilities.get_exposure_and_restriction_ages(exposure, entity)

    data.drop("age_group_id", axis=1, inplace=True)
    df = []
    for age_id in valid_ages:
        copied = data.copy()
        copied["age_group_id"] = age_id
        df.append(copied)
    data = pd.concat(df)
    data = utilities.normalize(data, fill_value=0, cols_to_fill=DISTRIBUTION_COLUMNS)
    data = data.filter(DEMOGRAPHIC_COLUMNS + DISTRIBUTION_COLUMNS)
    data = utilities.wide_to_long(data, DISTRIBUTION_COLUMNS, var_name="parameter")
    return data


def filter_relative_risk_to_cause_restrictions(data: pd.DataFrame) -> pd.DataFrame:
    """It applies age restrictions according to affected causes
    and affected measures. If affected measure is incidence_rate,
    it applies the yld_age_restrictions. If affected measure is
    excess_mortality_rate, it applies the yll_age_restrictions to filter
    the relative_risk data"""

    causes_map = {c.name: c for c in causes}
    temp = []
    affected_entities = set(data.affected_entity)
    affected_measures = set(data.affected_measure)
    for cause, measure in product(affected_entities, affected_measures):
        df = data[(data.affected_entity == cause) & (data.affected_measure == measure)]
        cause = causes_map[cause]
        if measure == "excess_mortality_rate":
            start, end = utilities.get_age_group_ids_by_restriction(cause, "yll")
        else:  # incidence_rate
            start, end = utilities.get_age_group_ids_by_restriction(cause, "yld")
        temp.append(df[df.age_group_id.isin(range(start, end + 1))])
    data = pd.concat(temp)
    return data


def get_relative_risk(entity: RiskFactor, location_id: int) -> pd.DataFrame:
    data = extract.extract_data(entity, "relative_risk", location_id)

    # FIXME: we don't currently support yll-only causes so I'm dropping them because the data in some cases is
    #  very messed up, with mort = morb = 1 (e.g., aortic aneurysm in the RR data for high systolic bp) -
    #  2/8/19 K.W.
    yll_only_causes = set([c.gbd_id for c in causes if c.restrictions.yll_only])
    data = data[~data.cause_id.isin(yll_only_causes)]

    data = utilities.convert_affected_entity(data, "cause_id")
    morbidity = data.morbidity == 1
    mortality = data.mortality == 1
    data.loc[morbidity & mortality, "affected_measure"] = "incidence_rate"
    data.loc[morbidity & ~mortality, "affected_measure"] = "incidence_rate"
    data.loc[~morbidity & mortality, "affected_measure"] = "excess_mortality_rate"
    data = filter_relative_risk_to_cause_restrictions(data)

    data = data.filter(
        DEMOGRAPHIC_COLUMNS
        + ["affected_entity", "affected_measure", "parameter"]
        + DRAW_COLUMNS
    )
    data = (
        data.groupby(["affected_entity", "parameter"])
        .apply(utilities.normalize, fill_value=1)
        .reset_index(drop=True)
    )

    if entity.distribution in ["dichotomous", "ordered_polytomous", "unordered_polytomous"]:
        tmrel_cat = utility_data.get_tmrel_category(entity)
        tmrel_mask = data.parameter == tmrel_cat
        data.loc[tmrel_mask, DRAW_COLUMNS] = data.loc[tmrel_mask, DRAW_COLUMNS].mask(
            np.isclose(data.loc[tmrel_mask, DRAW_COLUMNS], 1.0), 1.0
        )

    return data


def filter_by_relative_risk(df: pd.DataFrame, relative_risk: pd.DataFrame) -> pd.DataFrame:
    c_id = df.cause_id.unique()[0]
    rr = relative_risk[relative_risk.cause_id == c_id]
    #  We presume all attributable mortality moves through incidence.
    if set(rr.mortality) == {1} and set(rr.morbidity) == {1}:
        df = df[df.measure_id == MEASURES["YLDs"]]
    return df


def get_population_attributable_fraction(
    entity: Union[RiskFactor, Etiology], location_id: int
) -> pd.DataFrame:
    causes_map = {c.gbd_id: c for c in causes}
    if entity.kind == "risk_factor":
        data = extract.extract_data(entity, "population_attributable_fraction", location_id)
        relative_risk = extract.extract_data(entity, "relative_risk", location_id)

        # FIXME: we don't currently support yll-only causes so I'm dropping them because the data in some cases is
        #  very messed up, with mort = morb = 1 (e.g., aortic aneurysm in the RR data for high systolic bp) -
        #  2/8/19 K.W.
        yll_only_causes = set([c.gbd_id for c in causes if c.restrictions.yll_only])
        data = data[~data.cause_id.isin(yll_only_causes)]
        relative_risk = relative_risk[~relative_risk.cause_id.isin(yll_only_causes)]

        data = (
            data.groupby("cause_id", as_index=False)
            .apply(filter_by_relative_risk, relative_risk)
            .reset_index(drop=True)
        )

        temp = []
        # We filter paf age groups by cause level restrictions.
        for (c_id, measure), df in data.groupby(["cause_id", "measure_id"]):
            cause = causes_map[c_id]
            measure = "yll" if measure == MEASURES["YLLs"] else "yld"
            df = utilities.filter_data_by_restrictions(
                df, cause, measure, utility_data.get_age_group_ids()
            )
            temp.append(df)
        data = pd.concat(temp, ignore_index=True)

    else:  # etiology
        data = extract.extract_data(
            entity, "etiology_population_attributable_fraction", location_id
        )
        cause = [c for c in causes if entity in c.etiologies][0]
        data = utilities.filter_data_by_restrictions(
            data, cause, "inner", utility_data.get_age_group_ids()
        )
        if np.any(data[DRAW_COLUMNS] < 0):
            logger.warning(
                f"{entity.name.capitalize()} has negative values for paf. These will be replaced with 0."
            )
            other_cols = [c for c in data.columns if c not in DRAW_COLUMNS]
            data.set_index(other_cols, inplace=True)
            data = data.where(data[DRAW_COLUMNS] > 0, 0).reset_index()

    data = utilities.convert_affected_entity(data, "cause_id")
    data.loc[
        data["measure_id"] == MEASURES["YLLs"], "affected_measure"
    ] = "excess_mortality_rate"
    data.loc[data["measure_id"] == MEASURES["YLDs"], "affected_measure"] = "incidence_rate"
    data = (
        data.groupby(["affected_entity", "affected_measure"])
        .apply(utilities.normalize, fill_value=0)
        .reset_index(drop=True)
    )
    data = data.filter(
        DEMOGRAPHIC_COLUMNS + ["affected_entity", "affected_measure"] + DRAW_COLUMNS
    )
    return data


def get_estimate(entity: Covariate, location_id: int) -> pd.DataFrame:
    data = extract.extract_data(entity, "estimate", location_id)

    key_columns = ["location_id", "year_id"]
    if entity.by_age:
        key_columns.append("age_group_id")
    if entity.by_sex:
        key_columns.append("sex_id")

    data = data.filter(key_columns + COVARIATE_VALUE_COLUMNS)
    data = utilities.normalize(data)
    data = utilities.wide_to_long(data, COVARIATE_VALUE_COLUMNS, var_name="parameter")
    return data


def get_utilization_rate(entity: HealthcareEntity, location_id: int) -> pd.DataFrame:
    data = extract.extract_data(entity, "utilization_rate", location_id)
    data = utilities.normalize(data, fill_value=0)
    data = data.filter(DEMOGRAPHIC_COLUMNS + DRAW_COLUMNS)
    return data


def get_structure(entity: Population, location_id: int) -> pd.DataFrame:
    data = extract.extract_data(entity, "structure", location_id)
    data = data.drop("run_id", axis="columns").rename(columns={"population": "value"})
    data = utilities.normalize(data)
    return data


def get_theoretical_minimum_risk_life_expectancy(
    entity: Population, location_id: int
) -> pd.DataFrame:
    data = extract.extract_data(
        entity, "theoretical_minimum_risk_life_expectancy", location_id
    )
    data = data.rename(columns={"age": "age_start", "life_expectancy": "value"})
    data["age_end"] = data.age_start.shift(-1).fillna(125.0)
    return data


def get_age_bins(entity: Population, location_id: int) -> pd.DataFrame:
    age_bins = utility_data.get_age_bins()[["age_group_name", "age_start", "age_end"]]
    return age_bins


def get_demographic_dimensions(entity: Population, location_id: int) -> pd.DataFrame:
    demographic_dimensions = utility_data.get_demographic_dimensions(location_id)
    demographic_dimensions = utilities.normalize(demographic_dimensions)
    return demographic_dimensions



import pandas as pd
import pytest

from vivarium_inputs import utilities


@pytest.mark.parametrize(
    "sex_ids",
    [(1, 1, 1, 2, 2, 2), (1, 1, 2, 2, 3, 3), (1, 1, 1), (2, 2, 2), (3, 3, 3)],
    ids=["male_female", "male_female_both", "male", "female", "both"],
)
def test_normalize_sex(sex_ids):
    df = pd.DataFrame({"sex_id": sex_ids, "value": [1] * len(sex_ids)})
    normalized = utilities.normalize_sex(df, fill_value=0.0, cols_to_fill=["value"])
    assert {1, 2} == set(normalized.sex_id)


def test_normalize_sex_copy_3():
    values = [1, 2, 3, 4]
    df = pd.DataFrame({"sex_id": [3] * len(values), "value": values})
    normalized = utilities.normalize_sex(df, fill_value=0.0, cols_to_fill=["value"])
    assert (normalized.loc[normalized.sex_id == 1, "value"] == values).all()
    assert (normalized.loc[normalized.sex_id == 2, "value"] == values).all()


def test_normalize_sex_fill_value():
    values = [1, 2, 3, 4]
    fill = 0.0
    for sex in [1, 2]:
        missing_sex = 1 if sex == 2 else 2
        df = pd.DataFrame({"sex_id": [sex] * len(values), "value": values})
        normalized = utilities.normalize_sex(df, fill_value=fill, cols_to_fill=["value"])
        assert (normalized.loc[normalized.sex_id == sex, "value"] == values).all()
        assert (
            normalized.loc[normalized.sex_id == missing_sex, "value"] == [fill] * len(values)
        ).all()


def test_normalize_sex_no_sex_id():
    df = pd.DataFrame({"ColumnA": [1, 2, 3], "ColumnB": [1, 2, 3]})
    normalized = utilities.normalize_sex(df, fill_value=0.0, cols_to_fill=["value"])
    pd.testing.assert_frame_equal(df, normalized)
